@@ -749,6 +749,14 @@ Vector<ValueType>::create_real_view()
 }
 
 
+template <typename ValueType>
+std::unique_ptr<Vector<ValueType>> Vector<ValueType>::create_submatrix(
+    local_span rows, local_span columns, dim<2> global_size)
+{
+    return this->create_submatrix_impl(rows, columns, global_size);
+}
+
+
 template <typename ValueType>
 std::unique_ptr<Vector<ValueType>> Vector<ValueType>::create_with_same_config()
     const
@@ -769,6 +777,15 @@ std::unique_ptr<Vector<ValueType>> Vector<ValueType>::create_with_type_of_impl(
 }
 
 
+template <typename ValueType>
+std::unique_ptr<Vector<ValueType>> Vector<ValueType>::create_submatrix_impl(
+    local_span rows, local_span columns, dim<2> global_size)
+{
+    return Vector::create(this->get_executor(), this->get_communicator(),
+                          global_size, local_.create_submatrix(rows, columns));
+}
+
+
 #define GKO_DECLARE_DISTRIBUTED_VECTOR(ValueType) class Vector<ValueType>
 GKO_INSTANTIATE_FOR_EACH_VALUE_TYPE(GKO_DECLARE_DISTRIBUTED_VECTOR);
 
