@@ -1,46 +1,16 @@
-/*******************************<GINKGO LICENSE>******************************
-Copyright (c) 2017-2023, the Ginkgo authors
-All rights reserved.
-
-Redistribution and use in source and binary forms, with or without
-modification, are permitted provided that the following conditions
-are met:
-
-1. Redistributions of source code must retain the above copyright
-notice, this list of conditions and the following disclaimer.
-
-2. Redistributions in binary form must reproduce the above copyright
-notice, this list of conditions and the following disclaimer in the
-documentation and/or other materials provided with the distribution.
-
-3. Neither the name of the copyright holder nor the names of its
-contributors may be used to endorse or promote products derived from
-this software without specific prior written permission.
-
-THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
-IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
-TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
-PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
-HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
-SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
-LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
-DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
-THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
-(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
-OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
-******************************<GINKGO LICENSE>*******************************/
+// SPDX-FileCopyrightText: 2017 - 2025 The Ginkgo authors
+//
+// SPDX-License-Identifier: BSD-3-Clause
 
 #ifndef GKO_CORE_MATRIX_CSR_KERNELS_HPP_
 #define GKO_CORE_MATRIX_CSR_KERNELS_HPP_
 
 
-#include <ginkgo/core/matrix/csr.hpp>
-
-
 #include <ginkgo/core/base/array.hpp>
 #include <ginkgo/core/base/index_set.hpp>
 #include <ginkgo/core/base/types.hpp>
 #include <ginkgo/core/matrix/coo.hpp>
+#include <ginkgo/core/matrix/csr.hpp>
 #include <ginkgo/core/matrix/dense.hpp>
 #include <ginkgo/core/matrix/diagonal.hpp>
 #include <ginkgo/core/matrix/ell.hpp>
@@ -48,7 +18,6 @@ OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 #include <ginkgo/core/matrix/sellp.hpp>
 #include <ginkgo/core/matrix/sparsity_csr.hpp>
 
-
 #include "core/base/kernel_declaration.hpp"
 #include "core/matrix/csr_lookup.hpp"
 
@@ -88,6 +57,24 @@ namespace kernels {
                          const matrix::Csr<ValueType, IndexType>* d,  \
                          matrix::Csr<ValueType, IndexType>* c)
 
+#define GKO_DECLARE_CSR_SPGEMM_REUSE_KERNEL(ValueType, IndexType)          \
+    void spgemm_reuse(std::shared_ptr<const DefaultExecutor> exec,         \
+                      const matrix::Csr<ValueType, IndexType>* a,          \
+                      const matrix::Csr<ValueType, IndexType>* b,          \
+                      const matrix::csr::lookup_data<IndexType>& c_lookup, \
+                      matrix::Csr<ValueType, IndexType>* c)
+
+#define GKO_DECLARE_CSR_ADVANCED_SPGEMM_REUSE_KERNEL(ValueType, IndexType) \
+    void advanced_spgemm_reuse(                                            \
+        std::shared_ptr<const DefaultExecutor> exec,                       \
+        const matrix::Dense<ValueType>* alpha,                             \
+        const matrix::Csr<ValueType, IndexType>* a,                        \
+        const matrix::Csr<ValueType, IndexType>* b,                        \
+        const matrix::Dense<ValueType>* beta,                              \
+        const matrix::Csr<ValueType, IndexType>* d,                        \
+        const matrix::csr::lookup_data<IndexType>& c_lookup,               \
+        matrix::Csr<ValueType, IndexType>* c)
+
 #define GKO_DECLARE_CSR_SPGEAM_KERNEL(ValueType, IndexType)  \
     void spgeam(std::shared_ptr<const DefaultExecutor> exec, \
                 const matrix::Dense<ValueType>* alpha,       \
@@ -96,6 +83,14 @@ namespace kernels {
                 const matrix::Csr<ValueType, IndexType>* b,  \
                 matrix::Csr<ValueType, IndexType>* c)
 
+#define GKO_DECLARE_CSR_SPGEAM_NUMERIC_KERNEL(ValueType, IndexType)  \
+    void spgeam_numeric(std::shared_ptr<const DefaultExecutor> exec, \
+                        const matrix::Dense<ValueType>* alpha,       \
+                        const matrix::Csr<ValueType, IndexType>* a,  \
+                        const matrix::Dense<ValueType>* beta,        \
+                        const matrix::Csr<ValueType, IndexType>* b,  \
+                        matrix::Csr<ValueType, IndexType>* c)
+
 #define GKO_DECLARE_CSR_FILL_IN_DENSE_KERNEL(ValueType, IndexType)      \
     void fill_in_dense(std::shared_ptr<const DefaultExecutor> exec,     \
                        const matrix::Csr<ValueType, IndexType>* source, \
@@ -289,6 +284,12 @@ namespace kernels {
                           IndexType sample_size, IndexType* result)
 
 
+#define GKO_DECLARE_CSR_ROW_WISE_ABSOLUTE_SUM(ValueType, IndexType)           \
+    void row_wise_absolute_sum(std::shared_ptr<const DefaultExecutor> exec,   \
+                               const matrix::Csr<ValueType, IndexType>* orig, \
+                               array<ValueType>& sum)
+
+
 #define GKO_DECLARE_ALL_AS_TEMPLATES                                        \
     template <typename MatrixValueType, typename InputValueType,            \
               typename OutputValueType, typename IndexType>                 \
@@ -303,6 +304,12 @@ namespace kernels {
     template <typename ValueType, typename IndexType>                       \
     GKO_DECLARE_CSR_ADVANCED_SPGEMM_KERNEL(ValueType, IndexType);           \
     template <typename ValueType, typename IndexType>                       \
+    GKO_DECLARE_CSR_SPGEMM_REUSE_KERNEL(ValueType, IndexType);              \
+    template <typename ValueType, typename IndexType>                       \
+    GKO_DECLARE_CSR_ADVANCED_SPGEMM_REUSE_KERNEL(ValueType, IndexType);     \
+    template <typename ValueType, typename IndexType>                       \
+    GKO_DECLARE_CSR_SPGEAM_NUMERIC_KERNEL(ValueType, IndexType);            \
+    template <typename ValueType, typename IndexType>                       \
     GKO_DECLARE_CSR_SPGEAM_KERNEL(ValueType, IndexType);                    \
     template <typename ValueType, typename IndexType>                       \
     GKO_DECLARE_CSR_FILL_IN_DENSE_KERNEL(ValueType, IndexType);             \
@@ -367,7 +374,9 @@ namespace kernels {
     template <typename IndexType>                                           \
     GKO_DECLARE_CSR_BUILD_LOOKUP_KERNEL(IndexType);                         \
     template <typename IndexType>                                           \
-    GKO_DECLARE_CSR_BENCHMARK_LOOKUP_KERNEL(IndexType)
+    GKO_DECLARE_CSR_BENCHMARK_LOOKUP_KERNEL(IndexType);                     \
+    template <typename ValueType, typename IndexType>                       \
+    GKO_DECLARE_CSR_ROW_WISE_ABSOLUTE_SUM(ValueType, IndexType)
 
 
 GKO_DECLARE_FOR_ALL_EXECUTOR_NAMESPACES(csr, GKO_DECLARE_ALL_AS_TEMPLATES);
