@@ -1,9 +1,10 @@
-// SPDX-FileCopyrightText: 2017 - 2024 The Ginkgo authors
+// SPDX-FileCopyrightText: 2017 - 2025 The Ginkgo authors
 //
 // SPDX-License-Identifier: BSD-3-Clause
 
 #include "ginkgo/core/solver/cb_gmres.hpp"
 
+#include <string>
 #include <type_traits>
 
 #include <ginkgo/core/base/array.hpp>
@@ -17,6 +18,7 @@
 #include <ginkgo/core/matrix/dense.hpp>
 
 #include "core/base/extended_float.hpp"
+#include "core/config/config_helper.hpp"
 #include "core/config/solver_config.hpp"
 #include "core/solver/cb_gmres_accessor.hpp"
 #include "core/solver/cb_gmres_kernels.hpp"
@@ -162,11 +164,12 @@ typename CbGmres<ValueType>::parameters_type CbGmres<ValueType>::parse(
     const config::type_descriptor& td_for_child)
 {
     auto params = solver::CbGmres<ValueType>::build();
-    common_solver_parse(params, config, context, td_for_child);
-    if (auto& obj = config.get("krylov_dim")) {
+    config::config_check_decorator config_check(config);
+    config::common_solver_parse(params, config_check, context, td_for_child);
+    if (auto& obj = config_check.get("krylov_dim")) {
         params.with_krylov_dim(gko::config::get_value<size_type>(obj));
     }
-    if (auto& obj = config.get("storage_precision")) {
+    if (auto& obj = config_check.get("storage_precision")) {
         auto get_storage_precision = [](std::string str) {
             using gko::solver::cb_gmres::storage_precision;
             if (str == "keep") {
@@ -186,6 +189,7 @@ typename CbGmres<ValueType>::parameters_type CbGmres<ValueType>::parse(
         };
         params.with_storage_precision(get_storage_precision(obj.get_string()));
     }
+
     return params;
 }
 
